#! /bin/bash
#
#  Last modified:  January 17, 2020
#
#  This script must be executed in a directory that contains the grading harness
#  tar file (BSTGrader.tar) posted on the course website; that tar file must
#  contain the files:
#
#     c01driver.c
#     BST.h
#     Payload.h
#     PayloadWrapper.h
#     Monk.h
#     Monk.o
#
#  The directory must also contain a tar file containing your completed files:
#
#     Payload.c
#     PayloadWrapper.c
#     BST.c
#
#  We recommend that  you name that file yourPID.tar, using your VT email PID as
#  the first part of the file name.
#
#  Invocation:  ./gradec01.sh <name of your tar file>
#               e.g., ./gradec01.sh wmcquain.tar
#
#  There  will be a separate log file for each test phase that was completed, and
#  an overall report, named PID.txt if you followed the suggested naming convention.
#
#  Alternate invocation:  ./gradec01.sh -clean
#
#  This will remove the files created by an earlier run of the grading script;
#  it's useful if you just want to restart with a pristine environment.
#
#
#  Control whether student's source code submission is appended to the report:
appendSource="no"
#
#  Name of grading package file
gradingTar="BSTGrader.tar"

#  Names for directories
buildDir="build"
tempDir="tempFiles"

#   Names for log files created by this script:
buildLog="buildLog.txt"
testLog="testLog.txt"
seedValue="1506282647"   # need to modify c01driver.c to use wired seed value

#   Name for the executable will be set to PID_c01 later
exeName="c01"

#   Delimiter to separate sections of report file:
Separator="======================================================================"

#   Set results file names; conforms to test driver code
scoresLog="summaryLog.txt"

############################################# fn to check for tar file
#                 param1:  name of file to be checked
isTar() {

   mimeType=`file -b --mime-type $1`
   if [[ $mimeType == "application/x-tar" ]]; then
      return 0
   fi
   if [[ $mimeType == "application/x-gzip" ]]; then
      return 0
   fi
   return 1
}

##################################### fn to extract token from file name
#                 param1: (possibly fully-qualified) name of file
#  Note:  in production use, the first part of the file name will be the
#         student's PID
#
getPID() { 

   fname=$1
   # strip off any leading path info
   fname=${fname##*/}
   # extract first token of file name
   sPID=${fname%%.*}
}

###################################### fn to append a file to another file
#                        param1:  name of file to be appended
#                        param2:  name of file to be appended to
appendFile() {
   
   toAppend=$1
   toExtend=$2
   
   if [[ -e $toAppend ]]; then
      if [[ -e $toExtend ]]; then
         echo "$toAppend:" >> $toExtend
         echo >> $toExtend
         cat $toAppend >> $toExtend
         echo $Separator >> $toExtend
         return 0
      else
         echo "$toAppend was not found" >> $detailsLog
         return 1
      fi
   else
      echo "$toExtend was not found" >> $detailsLog
      return 1
   fi
   return 0
}
   
############################################################ Validate command line

# Verify number of parameters
   if [[ $# -ne 1 ]]; then
      echo "You must specify the name of your tar file (or -clean)."
      exit 1
   fi
   
# See if user selected the cleaning option
  if [[ $1 == "-clean" ]]; then
     echo "Removing earlier test files..."
     rm -Rf *.txt $buildDir $tempDir $exeName *_c01
     exit 0
  fi

# Verify presence  and type of named file
   sourceFile="$1"
   sourceFile=${sourceFile##*/}
   if [[ ! -e $sourceFile ]]; then
      echo "$sourceFile does not exist."
      exit 2
   fi
   isTar $sourceFile
   if [[ $? -ne 0 ]]; then
      echo "$sourceFile does not appear to be a valid tar file"
      exit 2
   fi
   
############################################################ Check for grading package

   if [[ ! -e $gradingTar ]]; then
      echo "$gradingTar is not present"
      exit 3
   fi
   
   isTar $gradingTar
   if [[ $? -ne 0 ]]; then
      echo "$gradingTar does not appear to be a tar file"
      exit 4
   fi

############################################################ Get student's PID
   
   # Extract first token of C file name (student PID when we run this)
   getPID $sourceFile
   
   # Reset name for executable file
   exeName=${sPID}_${exeName}
   
   # Initiate header for grading log
   headerLog="header.txt"
   echo "Grading:  $1" > $headerLog
   echo -n "Time:     " >> $headerLog
   echo `date` >> $headerLog
   echo >> $headerLog
   
############################################################ Prepare for build
  
   # Create log file:
   echo "Executing gradec1.sh..." > $buildLog
   echo >> $buildLog
   
   # Create build directory:
   echo "Creating build subdirectory" >> $buildLog
   echo >> $buildLog
   # Create build directory if needed; empty it if it already exists
   if [[ -d $buildDir ]]; then
      rm -Rf ./$buildDir/*
   else
      mkdir $buildDir
   fi
   
   # Unpack student's tar file to a temp directory
   if [[ ! -e $tempDir ]]; then
      mkdir "./$tempDir"
   fi
   
   echo "Copying student source file to $tempDir:" >> $buildLog
   tar xvf $1 -C $tempDir >> $buildLog
   
   # Verify and copy the three mandatory files into the build directory:
   mFile1="Payload.c"
   mFile2="PayloadWrapper.c"
   mFile3="BST.c"
   
   if [[ -e "$tempDir/$mFile1" ]]; then
      cp "$tempDir/$mFile1" ./$buildDir
   else
      echo "Student tar does not contain $mFile1"
      exit 5
   fi
   if [[ -e "$tempDir/$mFile2" ]]; then
      cp "$tempDir/$mFile2" ./$buildDir
   else
      echo "Student tar does not contain $mFile2"
      exit 5
   fi
   if [[ -e "$tempDir/$mFile3" ]]; then
      cp "$tempDir/$mFile3" ./$buildDir
   else
      echo "Student tar does not contain $mFile3"
      exit 5
   fi
   echo >> $buildLog
   
   # Unpack the test source files into the build directory
   echo "Unpacking test code into build directory:" >> $buildLog
   tar xf $gradingTar -C ./$buildDir >> $buildLog

   # Move to build directory
   cd ./$buildDir
   
############################################################ Build the test driver

   # build the driver; save output in build log
   echo "Compiling test code and submission" >> ../$buildLog
   gcc -o $exeName -std=c11 -Wall -ggdb3 c01driver.c BST.c Payload.c PayloadWrapper.c Monk.o >> ../$buildLog 2>&1
   echo >> ../$buildLog

   # Verify existence of executable
   if [[ ! -e $exeName ]]; then
      echo "Build failed; the file $exeName does not exist" >> ../$buildLog
      echo $Separator >> ../$buildLog
      mv ../$buildLog ../$sPID.txt
      exit 7
   fi
   if [[ ! -x $exeName ]]; then
      echo "Permissions error; the file $exeName is not executable" >> ../$buildLog
      echo $Separator >> ../$buildLog
      mv ../$buildLog ../$sPID.txt
      exit 8
   fi

   echo "Build succeeded..." >> $buildLog
   
   # Move executable up to test directory and return there
   echo "Moving the executable $exeName to the test directory." >> $buildLog
   mv ./$exeName .. >> $buildLog
   cd .. >> $buildLog
   
   # Delimit this section of the report file:
   echo $Separator >> $buildLog

############################################################ Perform testing

   # Initiate test Log
   echo -n "Begin testing... using seed: " > $testLog
   
   # The following sets the seed to the current time; comment out if you
   # want to use a hardwired seed for random data generation:
   seedValue=`date +%s`
   echo $seedValue >> $testLog
   echo >> $testLog
   
   # Execute the test code
   killed="no"

   timeout 30 ./$exeName -seed $seedValue >> $testLog 2>&1
   if [[ $? -eq 124 ]]; then
      echo "Testing your BST took more than 30 seconds, and was killed." >> $testLog
      killed="yes"
   fi

   # Accumulate individual test logs into details log
   detailsLog="details.txt"
   echo "Detailed results from tests:" > $detailsLog
   appendFile "initLog.txt" $detailsLog
   appendFile "insertLog.txt" $detailsLog
   appendFile "findLog.txt" $detailsLog
   
   echo >> $detailsLog

############################################################ Run valgrind check

   if [[ $killed = "no" ]]; then
	   #   full valgrind output is in $vgrindLog
	   #   extracted counts are in $vgrindIssues
	   vgrindLog="vgrindLog.txt"
	   echo "Running valgrind test..." >> $vgrindLog
	   vgrindSwitches=" --leak-check=full --show-leak-kinds=all --log-file=$vgrindLog --track-origins=yes -v"
	   scoreResultsLog2="ScoresValgrind.txt"
	   tmpVgrind="tmpVgrind.txt"
	   valgrind $vgrindSwitches ./$exeName -seed $seedValue >> $tmpVgrind 2>&1
	   if [[ -s $tmpVgrind ]]; then
	      cat $tmpVgrind >> $vgrindLog
	   fi
	   
	   # accumulate valgrind error counts
	   if [[ -e $vgrindLog ]]; then
	      vgrindIssues="vgrind_issues.txt"
	      echo "Valgrind issues:" > $vgrindIssues
	      grep "in use at exit" $vgrindLog >> $vgrindIssues
	      grep "total heap usage" $vgrindLog >> $vgrindIssues
	      grep "definitely lost" $vgrindLog >> $vgrindIssues
	      echo "Invalid reads: `grep -c "Invalid read" $vgrindLog`" >> $vgrindIssues
	      echo "Invalid writes: `grep -c "Invalid write" $vgrindLog`" >> $vgrindIssues
	      echo "Uses of uninitialized values: `grep -c "uninitialised" $vgrindLog`" >> $vgrindIssues
	   else
	      echo "Error running Valgrind test." >> $testLog
	   fi
   fi
############################################################ File report
# complete summary file

   summaryLog="$sPID.txt"
   
   # write header to summary log
   cat "$headerLog" > $summaryLog
   
   echo "Summary results from testing:" >> $summaryLog
   # write scores to summary file
   echo >> $summaryLog
   echo "Scores from testing" >> $summaryLog
   cat $scoresLog >> $summaryLog
   echo $Separator >> $summaryLog
   echo >> $summaryLog
   
   if [[ $killed = "no" ]]; then
	   # write Valgrind summary into log
	   echo "Summary of valgrind results:" >> $summaryLog
	   echo >> $summaryLog
	   cat $vgrindIssues >> $summaryLog
	   echo $Separator >> $summaryLog
   fi
   
   # write testing details into summary
   cat $testLog >> $summaryLog
   echo >> $summaryLog
   cat $detailsLog >> $summaryLog
   echo >> $summaryLog
   echo $Separator >> $summaryLog
   
   if [[ $killed = "no" ]]; then
	   # write Valgrind details into log
	   echo "Details from valgrind check:" >> $summaryLog
	   echo >> $summaryLog
	   cat $vgrindLog >> $summaryLog
	   echo $Separator >> $summaryLog
   fi

   # write build log into summary
   echo "Results from $buildLog" >> $summaryLog
   cat $buildLog >> $summaryLog
   echo >> $summaryLog
   
   # optionally append student's submission
   if [[ $appendSource == "yes" ]]; then
      echo "Submitted source code:" >> $summaryLog
      echo >> $summaryLog
      echo "$mFile1:" >> $summaryLog
      echo >> $summaryLog
      appendFile "./$buildDir/$mFile1" >> $summaryLog

      echo >> $summaryLog
      echo "$mFile2:" >> $summaryLog
      echo >> $summaryLog
      appendFile "./$buildDir/$mFile2" >> $summaryLog

      echo >> $summaryLog
      echo "$mFile3:" >> $summaryLog
      echo >> $summaryLog
      appendFile "./$buildDir/$mFile3" >> $summaryLog
   fi

exit 0
